/**
 * JASMINe
 * Copyright (C) 2011 BULL SA
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jasmine.monitoring.eos.monitoring.service;

import org.ow2.jasmine.event.beans.JasmineEventNotification;
import org.ow2.jasmine.event.messages.JasmineEvent;
import org.ow2.jasmine.event.messages.JasmineEventDetails;
import org.ow2.jasmine.monitoring.eos.monitoring.EventData;
import org.ow2.jasmine.monitoring.eos.monitoring.Serie;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

import javax.jms.*;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.*;

/**
 * This class is responsible for collecting data for a specific dashboard.
 * Data maybe collected from a file or database (replay mode) or from
 * the "jasmine" JMS Topic (live monitoring).
 *
 * @author durieuxp
 */
public class DashboardDataCollector implements MessageListener {

    /**
     * Logger.
     */
    private final static Log logger = LogFactory.getLog(MonitoringService.class);

    /**
     * Unique Dashboard Identifier
     */
    private int ident;

    /**
     * List of Series we are interested in
     */
    private List<Serie> series;   

    public static final String CONNECTION_FACTORY = "JCF";

    /**
     * List of Results (for all series)
     */
    LinkedList<EventData> resultList = null;

    /**
     * Max number of values returned to the client in a single call
     */
    private final static int MAX_RESULT_READ = 5000;

    public static final String RESULT_TOPIC = "jasmine";
    private Topic resultTopic = null;
    private Connection resultConnection = null;
    private Session resultSession = null;
    private MessageConsumer resultSubscriber = null;

    /**
     * List of notifications
     */
    LinkedList<JasmineEventNotification> notifList = null;

    public static final String NOTIF_TOPIC = "jasmineNotification";
    private Topic notifTopic = null;
    private Connection notifConnection = null;
    private Session notifSession = null;
    private MessageConsumer notifSubscriber = null;

    private Extractor extractor = null;

    /**
     * Mode is replay (true) or live monitoring (false)
     */
    private boolean replay;

    // ---------------------------------------------------------------------
    // Constructors
    // ---------------------------------------------------------------------

    public DashboardDataCollector(final int ident, final List<Serie> series, final boolean notifications) {
        this.ident = ident;
        this.series = series;
        cleanResults();
    }

    // ---------------------------------------------------------------------
    // Public methods used for Extractors
    // ---------------------------------------------------------------------

    /**
     * Get the list of Series
     */
    public List<Serie> getSeries() {
        return series;
    }

    /**
     * Add a Result to the list
     * @param evd  result
     */
    public synchronized void addResult(EventData evd) {
        resultList.addLast(evd);
    }

    public void setReplayCompleted() {
        logger.debug("");
        extractor = null;
    }

    /**
     * Check if an objectname mbean matches a JMX pattern
     * @param strmbean
     * @param strpattern
     * @return  true if matches
     */
    public boolean match(final String strmbean, final String strpattern) {
        if (strpattern == null || strpattern.length() == 0 || strpattern.equals("*:*")) {
            // No pattern defined.
            return true;
        }
        //logger.debug("compare pattern:" + strpattern + " to mbean:" + strmbean);
        ObjectName mbean = null;
        ObjectName pattern = null;
        try {
            mbean = new ObjectName(strmbean);
        } catch (MalformedObjectNameException e) {
            logger.error("Bad MBean ObjectName (MalformedObjectNameException) " + strmbean);
            return false;
        } catch (NullPointerException e) {
            logger.error("Bad MBean ObjectName (NPE) " + strmbean);
            return false;
        }
        try {
            pattern = new ObjectName(strpattern);
        } catch (MalformedObjectNameException e) {
            logger.error("Bad Pattern ObjectName (MalformedObjectNameException) " + strpattern);
            return false;
        } catch (NullPointerException e) {
            logger.error("Bad Pattern ObjectName (NPE) " + strpattern);
            return false;
        }
        if (pattern == null) {
            logger.error("Cannot get ObjectName on " + strpattern);
            return false;
        }

        return pattern.apply(mbean);
    }

    // ---------------------------------------------------------------------
    // MessageListener implementation
    // ---------------------------------------------------------------------

    /**
     * Used for live monitoring
     * A message has been read from one of the topics
     * @param message
     */
    public void onMessage(Message message) {

        // deal only with ObjectMessage
        if (!(message instanceof ObjectMessage)) {
            return;
        }
        ObjectMessage om = (ObjectMessage) message;

        // We are interested only in JasmineEvent messages
        Object obj = null;
        try {
            message.acknowledge();
            obj = om.getObject();
        } catch (JMSException e) {
            logger.error("Bad message got on Jms Topic");
            return;
        }
        if ((obj instanceof JasmineEvent)) {
            resultReceived((JasmineEvent) obj);
        } else if (obj instanceof JasmineEventNotification) {
            notifReceived((JasmineEventNotification) obj);
        }
    }

    private void resultReceived(JasmineEvent event) {
        logger.debug("");

        // Each JasmineEvent may contains more than 1 value.
        String server = event.getServer();
        for (JasmineEventDetails ed : event.getEvents()) {
            String probe = ed.getProbe();
            //logger.debug("Mbean:Attribute = " + probe);
            int index = probe.lastIndexOf(":");
            String mbean = probe.substring(0, index);
            String attr = probe.substring(index + 1);

            // Find series accepting this event.
            for (Serie curser : series) {
                //logger.debug("look if matches serie:" + curser.getId());

                // Check Attribute
                if (! attrmatch(attr, curser.getYcol())) {
                    continue;
                }

                // Check if pattern matches the mbean
                if (!match(mbean, curser.getPattern())) {
                    //logger.debug("Mbean do not match :" + curser.getId());
                    continue;
                }

                // Check server
                String svr = curser.getServer();
                if (svr != null && svr.length() > 0 && !svr.equals(server)) {
                    //logger.debug("Server do not match :" + curser.getId());
                    continue;
                }

                // This event matches this Serie and will be kept
                EventData evd = new EventData(curser.getId(), ed.getTimestamp().getTime(), ed.getValue());
                //logger.debug("matches!");
                addResult(evd);
                // continue because an event may match several series
            }
        }
    }

    private void notifReceived(JasmineEventNotification notif) {
        logger.debug("");
        notifList.add(notif);
    }

    // ---------------------------------------------------------------------
    // Interface for the MonitoringService
    // ---------------------------------------------------------------------

    /**
     * Get all notifs
     */
    public synchronized List<JasmineEventNotification> getAllNotifs() {
        List<JasmineEventNotification> res = new LinkedList<JasmineEventNotification>();
        res.addAll(notifList);
        notifList = new LinkedList<JasmineEventNotification>();
        logger.debug("returning " + res.size() + " notifs");
        return res;
    }
    
    /**
     * Get all results collected and reset the internal list.
     * These results may be from replay, or from live monitoring.
     *
     * @return List of EventData, or null if replay is completed.
     */
    public synchronized List<EventData> getAllResults() {

        List<EventData> res = new ArrayList<EventData>();
        int count = 0;
        while (count < MAX_RESULT_READ) {
            try {
                EventData ed = resultList.removeFirst();
                res.add(ed);
                count++;
            } catch (NoSuchElementException e) {
                break;
            }
        }
        if (count == 0 && isReplayCompleted()) {
            logger.debug("ReplayCompleted");
            return null;
        }
        logger.debug("Found " + count + " events");
        return res;
    }

    /**
     * Start the live monitoring for this dashboard.
     */
    public void startLiveMonitoring() throws NamingException, JMSException {
        logger.debug("");

        cleanResults();

        replay = false;

        Context context = new InitialContext();
        ConnectionFactory tcf = (ConnectionFactory) context.lookup(CONNECTION_FACTORY);

        // Get the jasmine Topic and start listening on it
        if (resultTopic == null) {
            resultTopic = (Topic) context.lookup(RESULT_TOPIC);
        }
        resultConnection = tcf.createConnection();
        resultSession = resultConnection.createSession(false, Session.CLIENT_ACKNOWLEDGE);
        resultSubscriber = resultSession.createConsumer(resultTopic);
        resultSubscriber.setMessageListener(this);
        resultConnection.start();

        // Get the jasmineNotification Topic and start listening on it
        if (notifTopic == null) {
            notifTopic = (Topic) context.lookup(NOTIF_TOPIC);
        }
        notifConnection = tcf.createConnection();
        notifSession = notifConnection.createSession(false, Session.CLIENT_ACKNOWLEDGE);
        notifSubscriber = notifSession.createConsumer(notifTopic);
        notifSubscriber.setMessageListener(this);
        notifConnection.start();
    }

    /**
     * Stop the live monitoing for this dashboard.
     */
    public void stopLiveMonitoring() {
        logger.debug("");

        if (resultConnection != null) {
            try {
                if (resultSubscriber != null) {
                    resultSubscriber.close();
                }
                if (resultSession != null) {
                    resultSession.close();
                }
                resultConnection.close();
            } catch (JMSException e) {
                logger.error("cannot disconnect from jasmine topic");
            } finally {
                resultConnection = null;
                resultSession = null;
                resultSubscriber = null;
            }
        }
        if (notifConnection != null) {
            try {
                if (notifSubscriber != null) {
                    notifSubscriber.close();
                }
                if (notifSession != null) {
                    notifSession.close();
                }
                notifConnection.close();
            } catch (JMSException e) {
                logger.error("cannot disconnect from jasmineNotification topic");
            } finally {
                notifConnection = null;
                notifSession = null;
                notifSubscriber = null;
            }
        }

        if (extractor != null) {
            // Must stop the thread extracting data from file or database.
            extractor.stopMonitoring();
        }
    }

    /**
     * Start Replay from file
     * @param fileName
     * @param from
     * @param to
     * @throws Exception
     */
    public void startFileReplay(final String fileName, final Date from, final Date to) throws Exception {
        logger.debug("from="+from);

        cleanResults();

        // Check that File exists and is not empty
        FileReader f = new FileReader(fileName);
        BufferedReader buffer = new BufferedReader(f);
        buffer.mark(500);
        String header = buffer.readLine();
        if (header == null) {
            logger.warn("Empty file: " + fileName);
            throw new Exception("Empty file: " + fileName);
        }

        replay = true;

        // Create the Extractor
        if (header.startsWith("date")) {
            extractor = new HeaderFileExtractor(this, header, 0, buffer, from, to);
        } else if (header.startsWith("time")) {
            extractor = new HeaderFileExtractor(this, header, 1, buffer, from, to);
        } else {
            // reset buffer to get the first line
            buffer.reset();
            extractor = new FileExtractor(this, buffer, from, to);
        }

        // Start the Extractor to begin data collect
        extractor.start();
    }

    // ---------------------------------------------------------------------
    // replay from database
    // ---------------------------------------------------------------------

    /**
     * Start replay from database
     * @param from
     * @param to
     * @throws Exception
     */
    public void startDatabaseReplay(final Date from, final Date to) throws Exception {
        logger.debug("");

        cleanResults();
        replay = true;

        // Create the Extractor
        // Should be already there because created by getDatabaseLimits
        if (extractor == null) {
            extractor = new DatabaseExtractor(this);
        }

        // Start the Extractor to begin data collect
        extractor.setDateLimits(from, to);
        extractor.start();
    }

    /**
     * Get the oldest timestamp in the database. Should be the 1st record.
     * @return timestamp
     * @throws Exception
     */
    public long getDatabaseLimits() throws Exception {
        logger.debug("");

        // Create the Extractor
        extractor = new DatabaseExtractor(this);

        // create a temporary extractor to do the job.
        return ((DatabaseExtractor)extractor).getStartTime();
    }

    // ---------------------------------------------------------------------
    // Private methods
    // ---------------------------------------------------------------------

    /**
     * Check if replay is completed or not.
     * @return
     */
    private boolean isReplayCompleted() {
        return (replay && extractor == null);
    }

    /**
     * Clean datas
     */
    private synchronized void cleanResults() {
        resultList = new LinkedList<EventData>();
        notifList = new LinkedList<JasmineEventNotification>();
    }

    /**
     * Check if Attr matches the ycol field
     * @param attr Attribute found in the result
     * @param ycol Definition in the y.col field of configuration file.
     * @return true if matches
     */
    protected boolean attrmatch(String attr, String ycol) {
        if (attr.equals(ycol)) {
            // match exactly
            return true;
        }
        int index = attr.indexOf(".");
        if (index > 0 && attr.substring(index + 1).equals(ycol)) {
            // matches the attribute name
            // Configuration did not care about the indicator.
            return true;
        }
        //logger.debug("attr do not match:" + ycol);
        return false;
    }
}
