/**
 * JASMINe
 * Copyright (C) 2009 Bull S.A.S.
 * Copyright (C) 2010 France Telecom
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.monitoring.eos.monitoring.service;

import org.ow2.jasmine.monitoring.eos.monitoring.EventData;
import org.ow2.jasmine.monitoring.eos.monitoring.Serie;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.*;

/**
 * Extractor for File produced by JasmineProbe OR eventswitch
 * Both formats are supposed to be the same, as described here after:
 *
 * Fields are separated by ;
 * 0 = timestamp
 * 1 = date formatted
 * 2 = target
 * 3 = domain
 * 4 = server
 * 5 = mbean:attr
 * 6 = value
 *
 * @author durieuxp
 */
public class FileExtractor extends Extractor {

    private BufferedReader buffer;

    private long fromTimestamp = 0L;

    private long toTimestamp = 0L;

    // ------------------------------------------------------------------------
    // Constructors
    // ------------------------------------------------------------------------

    /**
     * Constructor for DashboardDataCollector
     */
    public FileExtractor(DashboardDataCollector dbc, BufferedReader buffer, Date from, Date to) {
        super(dbc);
        setDateLimits(from, to);
        this.buffer = buffer;
        if (from != null) {
            this.fromTimestamp = from.getTime();
        }
        if (to != null) {
            this.toTimestamp = to.getTime();
        }
    }

    // ------------------------------------------------------------------------
    // Thread implementation
    // ------------------------------------------------------------------------

    /**
     * Run method called when the thread is started.
     */
    public void run() {
        logger.debug("start reading file");

        // parse the file
        String line = "";
        long timestamp = 0;
        try {
            while (! stopWanted) {
                line = buffer.readLine();
                if (line == null) {
                    if (timestamp >= toTimestamp) {
                        logger.debug("End of file");
                        break;
                    } else {
                        logger.debug("Wait for new event");
                        Thread.sleep(5000);
                        continue;
                    }
                }
                //logger.debug("Reading line:" + line);
                // split the line just once
                String[] split = line.split(";");
                if (split.length <= 6) {
                    // Ignore all lines incomplete.
                    logger.debug("ignore incomplete line: " + line);
                    continue;
                }
                timestamp = new Long(split[0]);     // timestamp
                String date = split[1];     // formatted date
                String server = split[4];   // server name

                // Check date
                if (timestamp < fromTimestamp) {
                    // logger.debug("too old");
                    continue;
                }
                if (toTimestamp > 0 && timestamp > toTimestamp) {
                    logger.debug("ignoring the rest of the file : not in the time range");
                    break;
                }

                // compute mbean and attribute
                String mbeanattr = split[5];
                int index = mbeanattr.lastIndexOf(":");
                String mbean = mbeanattr.substring(0, index);
                String attr = mbeanattr.substring(index + 1);

                String value = split[6];

                // Find series accepting this event.
                for (Serie curser : dbc.getSeries()) {
                    // Check Attribute
                    if (! attrmatch(attr, curser.getYcol())) {
                        continue;
                    }

                    // Check pattern
                    if (! dbc.match(mbean, curser.getPattern())) {
                        continue;
                    }

                    // Check server
                    if (curser.getServer() != null && curser.getServer().length() > 0 && !curser.getServer().equals(server)) {
                        //logger.debug("bad server:" + curser.getServer());
                        continue;
                    }

                    // This event matches this Serie and will be kept
                    EventData ed = new EventData(curser.getId(), timestamp, value);
                    dbc.addResult(ed);
                    // continue because an event may match several series
                }
            }
        } catch (IOException e) {
            logger.error("Error reading file: " + e);
        } catch (InterruptedException e) {
            logger.warn("Interrupted");
        } finally {
            dbc.setReplayCompleted();
        }
    }
}
